import funcy as fn

from lazytree import LazyTree


def split(itvl):
    lo, hi = itvl
    mid = lo + (hi - lo)/2
    return (lo, mid), (mid, hi)


def test_children():
    tree = LazyTree(root=(0, 1), child_map=split)
    assert [c.view() for c in tree.children] == [(0, 0.5), (0.5, 1)]
    tree2 = tree.children[0]
    assert [c.view() for c in tree2.children] == [(0, 0.25), (0.25, .5)]


def test_map():
    tree = LazyTree(root=(0, 1), child_map=split, view=sum)
    tree2 = LazyTree(root=(0, 1), child_map=split).map(sum)
    assert tree.view() == tree2.view()

    for c1, c2 in zip(tree.children, tree2.children):
        assert c1.view() == c2.view()

    assert tree.with_identity_view().view() == (0, 1)


def assert_contracting(elems):
    max_size = 1
    for size in fn.take(7, elems):
        assert 0 <= size <= max_size
        max_size = size

    assert max_size < 0.5


def test_bfs():
    tree = LazyTree(root=(0, 1), child_map=split, view=lambda x: x[1] - x[0])
    assert_contracting(tree.bfs())


def test_iddfs():
    tree = LazyTree(root=(0, 1), child_map=split, view=lambda x: x[1] - x[0])
    nodes = list(tree.iddfs(max_depth=3))
    assert_contracting(nodes)
    assert list(tree.prune(isleaf=lambda s: s < 0.2).bfs()) == nodes


def test_cost_guided_traversal():
    tree = LazyTree(root=(0, 1), child_map=split, view=lambda x: x[1] - x[0])
    assert_contracting(tree.cost_guided_traversal(lambda x: -x))

    sizes1 = fn.take(5, tree.cost_guided_traversal(lambda x: x))
    sizes2 = fn.take(5, tree.bfs())
    assert sizes1 == sizes2


def test_prune_and_leaves():
    tree = LazyTree(root=(0, 1), child_map=split, view=lambda x: x[1] - x[0])
    tree2 = tree.prune(isleaf=lambda x: x <= 0.2)

    assert [c.view() for c in tree.children] == \
        [c.view() for c in tree2.children]

    assert all(c >= 0.1 for c in tree2.leaves())
    assert len(list(tree2.leaves())) == 8



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['lazytree']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.1,<20.0', 'funcy>=1.12,<2.0']

setup_kwargs = {
    'name': 'lazytree',
    'version': '0.3.0',
    'description': 'Python library for manipulating infinite trees.',
    'long_description': None,
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'marcell.vc@eecs.berkeley.edu',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from collections import deque
from heapq import heappush, heappop

import attr
import funcy as fn


@attr.s(frozen=True)
class LazyTree:
    root = attr.ib()
    child_map = attr.ib()
    _view = attr.ib(default=lambda x: x)
    _isleaf = attr.ib(default=lambda _: False)  # For pruning subtrees.

    def isleaf(self):
        return self._isleaf(self.view())

    @property
    def children(self):
        if self.isleaf():
            return ()

        _children = self.child_map(self.root)
        return tuple(attr.evolve(self, root=c) for c in _children)

    def view(self):
        return self._view(self.root)

    def map(self, func):
        return attr.evolve(self, view=fn.compose(func, self._view))

    def prune(self, isleaf):
        """Return a new LazyTree where child nodes satisfying
        isleaf have their subtrees pruned."""
        return attr.evolve(self, isleaf=lambda x: self._isleaf(x) or isleaf(x))

    def with_identity_view(self):
        return attr.evolve(self, view=lambda x: x)

    # Traversals

    def bfs(self):
        queue = deque([self])
        while len(queue) > 0:
            curr = queue.pop()
            yield curr.view()
            queue.extendleft(curr.children)

    def cost_guided_traversal(self, cost_map):
        queue = [(-cost_map(self.view()), self)]
        while len(queue) > 0:
            _, curr = heappop(queue)
            yield curr.view()
            for c in curr.children:
                cost = -cost_map(c.view())
                heappush(queue, (cost, c))

    def leaves(self, *, max_depth=float('inf')):
        stack = [(self, 0)]
        while len(stack) > 0:
            curr, depth = stack.pop()
            if len(curr.children) == 0 or depth >= max_depth:
                yield curr.view()
            else:
                depth += 1
                stack.extend([(c, depth) for c in curr.children])

    def iddfs(self, max_depth=float('inf'), flatten=True):
        """Iterative deepening depth-first search."""
        depth = 0
        while depth <= max_depth:
            nodes = self.leaves(max_depth=depth)
            if flatten:
                yield from nodes
            else:
                yield tuple(nodes)
            depth += 1
